import csv
import json
import math
import os
import re
from shlex import quote
import shutil
import textwrap
from datetime import timedelta
from pathlib import Path
from shutil import which

import humanize
import pandas as pd
import sqlite_utils
from rich import inspect
from rich.prompt import Confirm
from tabulate import tabulate

from xklb.subtitle import is_file_with_subtitle

from .db import sqlite_con
from .utils import (
    cmd,
    conditional_filter,
    get_ip_of_chromecast,
    get_ordinal_media,
    log,
    parse_args,
    print_query,
    remove_media,
    stop,
)

def play_mpv(args, video_path: Path):
    if which("mpv") is not None:
        player = ["mpv", "--fs", "--force-window=yes", "--terminal=no"]
    else:
        player = ["xdg-open"]

    if args.chromecast:
        subs = json.loads(
            cmd(
                "ffprobe",
                "-loglevel",
                "error",
                "-select_streams",
                "s",
                "-show_entries",
                "stream",
                "-of",
                "json",
                video_path,
            ).stdout
        )["streams"]

        subtitles_file = None
        if len(subs) > 0:
            db = sqlite_utils.Database(memory=True)
            db["subs"].insert_all(subs, pk="index")  # type: ignore
            subtitle_index = db.execute_returning_dicts(
                """select "index" from subs
                order by
                    lower(tags) like "%eng%" desc
                    , lower(tags) like "%dialog%" desc
                limit 1"""
            )[0]["index"]
            log.debug(f"Using subtitle {subtitle_index}")

            subtitles_file = cmd("mktemp --suffix=.vtt --dry-run").stdout.strip()
            cmd(
                f'ffmpeg -loglevel warning -txt_format text -i {video_path} -map "0:{subtitle_index}" "{subtitles_file}"'
            )

        if args.vlc:
            watched = cmd(
                f"vlc --sout '#chromecast' --sout-chromecast-ip={args.cc_ip} --demux-filter=demux_chromecast {'--sub-file='+subtitles_file if subtitles_file else ''} {video_path}"
            )
        else:
            watched = cmd(
                f"catt -d '{args.chromecast_device}' cast {video_path} {'--subtitles '+subtitles_file if subtitles_file else ''}"
            )

        if subtitles_file:
            Path(subtitles_file).unlink(missing_ok=True)

        if "Heartbeat timeout, resetting connection" in watched.stderr:
            raise Exception("Media is possibly partially unwatched")

        if watched.stderr == "":
            raise Exception("catt does not exit nonzero? but something might have gone wrong")

        return  # end of chromecast

    if not is_file_with_subtitle(video_path):
        player.append("--speed=1.7")
    else:
        player.append("--speed=1")

    cmd(*player, '--', video_path)


def keep_video(video: Path):
    kp = re.match(".*?/mnt/d/(.*?)/", str(video))
    if kp:
        keep_path = Path(kp[0], "keep/")
    else:
        keep_path = video.parent / "keep/"

    keep_path.mkdir(parents=True, exist_ok=True)
    shutil.move(video, keep_path)


def watch(args):
    con = sqlite_con(args.db)

    bindings = []
    filename_include_sql = ""
    filename_exclude_sql = ""
    if args.search:
        for inc in args.search.split(","):
            filename_include_sql += " AND filename LIKE ? "
            bindings.append("%" + inc.replace(" ", "%").replace("%%", " ") + "%")
    if args.exclude:
        for exc in args.exclude.split(","):
            filename_exclude_sql += " AND filename NOT LIKE ? "
            bindings.append("%" + exc.replace(" ", "%").replace("%%", " ") + "%")

    sql_filter = conditional_filter(args)

    LIMIT = "LIMIT " + str(args.limit)
    OFFSET = f"OFFSET {args.skip}" if args.skip else ""

    query = f"""
    SELECT filename, duration/60/60 as hours, duration / size AS seconds_per_byte, size
    FROM media
    WHERE 1=1
    {filename_include_sql}
    {filename_exclude_sql}
    {sql_filter}
    ORDER BY
            {args.sort + ',' if args.sort else ''}
            {'filename,' if args.print or args.search or args.play_in_order > 0 else ''}
            seconds_per_byte ASC
    {LIMIT} {OFFSET}
    """

    if args.print and "a" in args.print:
        query = f"""select
            "Aggregate" as filename
            , sum(hours) hours
            , avg(seconds_per_byte) seconds_per_byte
            , sum(size) size
        from ({query})
        """

    if args.print and "q" in args.print:
        print_query(bindings, query)
        if args.play_in_order > 1:
            get_ordinal_media(con, args, Path("vid"), sql_filter)
        stop()

    if args.chromecast:
        args.cc_ip = get_ip_of_chromecast(args.device_name)

    if args.print:
        videos = pd.DataFrame([dict(r) for r in con.execute(query, bindings).fetchall()])

        if "f" in args.print:
            if args.limit == 1:
                f = videos[["filename"]].loc[0].iat[0]
                if not Path(f).exists():
                    remove_media(con, f)
                    return main()
                print(f)
            else:
                csvf = videos[["filename"]].to_csv(index=False, header=False, sep="\t", quoting=csv.QUOTE_NONE)
                print(csvf.strip())
        else:
            table_content = videos
            table_content[["filename"]] = table_content[["filename"]].applymap(
                lambda x: textwrap.fill(x, os.get_terminal_size().columns - 30)
            )
            table_content[["size"]] = table_content[["size"]].applymap(lambda x: humanize.naturalsize(x))
            print(
                tabulate(
                    table_content[["filename", "size", "hours"]],
                    tablefmt="fancy_grid",
                    headers="keys",
                    showindex=False,
                )
            )
            summary = videos.sum(numeric_only=True)
            duration = timedelta(hours=int(summary.hours), minutes=math.ceil(summary.hours % 1 * 60))
            print("Total duration:", humanize.precisedelta(duration, minimum_unit="minutes"))

        stop()

    if args.move:
        Path(args.move).mkdir(exist_ok=True, parents=True)
        keep_path = str(Path(args.move).resolve())

        videos = pd.DataFrame([dict(r) for r in con.execute(query, bindings).fetchall()])
        for video in videos[["filename"]]:
            if Path(video).exists() and "/keep/" not in video:
                shutil.move(video, keep_path)

            remove_media(con, video)
        stop()

    next_video = con.execute(query, bindings).fetchone()
    if next_video is None:
        print('No media found')
        stop()

    next_video = dict(next_video)["filename"]
    if args.play_in_order > 1:
        next_video = get_ordinal_media(con, args, Path(next_video), sql_filter)

    original_video = Path(next_video)
    next_video = Path(args.prefix + next_video)
    print(next_video)

    if "/keep/" in str(next_video):
        keep_video(next_video)
        remove_media(con, original_video)
        exit()

    if next_video.exists():
        if args.only_video:
            has_video = (
                cmd(
                    f"ffprobe -show_streams -select_streams v -loglevel error -i {quote(str(next_video))} | wc -l",
                    quiet=True,
                    shell=True,
                ).stdout
                > "0"
            )
            if not has_video:
                remove_media(con, original_video)
                exit()

        if args.force_transcode:
            temp_video = cmd("mktemp", "--suffix=.mkv", "--dry-run").stdout.strip()
            shutil.move(next_video, temp_video)
            next_video = next_video.with_suffix(".mkv")
            cmd(
                (
                    f"ffmpeg -loglevel error -stats -i {temp_video} -map 0 -scodec webvtt -vcodec h264"
                    " -preset fast -profile:v high -level 4.1 -crf 17 -pix_fmt yuv420p"
                    " -acodec opus -ac 2 -b:a 128k -filter:a loudnorm=i=-18:lra=17"
                    f" {next_video} && rm {temp_video}"
                )
            )
            print(next_video)

        play_mpv(args, next_video)

        if args.action == "keep":
            keep_video(next_video)
        elif args.action == "ask" and Confirm.ask("Keep?", default=False):
            keep_video(next_video)
        else:
            if len(args.prefix) > 0:
                next_video.unlink()
                cmd("/bin/rm", next_video)
            else:
                cmd("trash-put", next_video)

    remove_media(con, original_video)


def main():
    args = parse_args(default_chromecast="Living Room TV")

    watch(args)


if __name__ == "__main__":
    main()



import argparse
import json
import math
import os
import sys
from datetime import datetime
from pathlib import Path
from typing import Dict

import mutagen
import pandas as pd
from joblib import Parallel, delayed
from rich import inspect
from tinytag import TinyTag

from .db import fetchall_dict, sqlite_con
from .subtitle import get_subtitle, is_file_with_subtitle, youtube_dl_id
from .utils import chunks, cmd, combine, get_video_files, log, remove_None, safe_unpack

SQLITE_PARAM_LIMIT = 32765


def get_provenance(file):
    if youtube_dl_id(file) != "":
        return "YouTube"

    return None


def parse_tags(mutagen: Dict, tinytag: Dict):
    tags = {
        "mood": combine(
            mutagen.get("albummood"),
            mutagen.get("MusicMatch_Situation"),
            mutagen.get("Songs-DB_Occasion"),
            mutagen.get("albumgrouping"),
        ),
        "genre": combine(mutagen.get("genre"), tinytag.get("genre"), mutagen.get("albumgenre")),
        "year": combine(
            mutagen.get("originalyear"),
            mutagen.get("TDOR"),
            mutagen.get("TORY"),
            mutagen.get("date"),
            mutagen.get("TDRC"),
            mutagen.get("TDRL"),
            tinytag.get("year"),
        ),
        "bpm": safe_unpack(mutagen.get("fBPM"), mutagen.get("bpm_accuracy")),
        "key": safe_unpack(mutagen.get("TIT1"), mutagen.get("key_accuracy"), mutagen.get("TKEY")),
        "time": combine(mutagen.get("time_signature")),
        "decade": safe_unpack(mutagen.get("Songs-DB_Custom1")),
        "categories": safe_unpack(mutagen.get("Songs-DB_Custom2")),
        "city": safe_unpack(mutagen.get("Songs-DB_Custom3")),
        "country": combine(
            mutagen.get("Songs-DB_Custom4"),
            mutagen.get("MusicBrainz Album Release Country"),
            mutagen.get("musicbrainz album release country"),
            mutagen.get("language"),
        ),
        "description": combine(
            mutagen.get("description"),
            mutagen.get("lyrics"),
            tinytag.get("comment"),
        ),
        "album": safe_unpack(tinytag.get("album"), mutagen.get("album")),
        "title": safe_unpack(tinytag.get("title"), mutagen.get("title")),
        "artist": combine(
            tinytag.get("artist"),
            mutagen.get("artist"),
            mutagen.get("artists"),
            tinytag.get("albumartist"),
            tinytag.get("composer"),
        ),
    }

    # print(mutagen)
    # breakpoint()

    return tags


def extract_metadata(args, f):
    try:
        probe = json.loads(
            cmd(
                "ffprobe",
                "-loglevel",
                "quiet",
                "-print_format",
                "json=compact=1",
                "-show_entries",
                "format",
                f,
                quiet=True,
            ).stdout
        )
    except (KeyboardInterrupt, SystemExit):
        exit(130)
    except:
        print(f"Failed reading {f}", file=sys.stderr)
        cmd("trash-put", f, strict=False)
        return
    if not "format" in probe:
        print(f"Failed reading format {f}", file=sys.stderr)
        print(probe)
        return

    stat = os.stat(f)
    blocks_allocated = stat.st_blocks * 512

    probe["format"].pop("tags", None)
    probe["format"].pop("format_long_name", None)
    probe["format"].pop("nb_programs", None)
    probe["format"].pop("nb_streams", None)
    probe["format"].pop("probe_score", None)
    probe["format"].pop("start_time", None)
    probe["format"].pop("probe_score", None)
    probe["format"].pop("probe_score", None)

    if "size" in probe["format"]:
        probe["format"]["size"] = int(probe["format"]["size"])

    if blocks_allocated == 0:
        sparseness = 0
    else:
        sparseness = probe["format"]["size"] / blocks_allocated

    media = dict(
        **probe["format"],
        # streams=probe["streams"],
        sparseness=sparseness,
        time_created=datetime.fromtimestamp(stat.st_ctime),
        time_modified=datetime.fromtimestamp(stat.st_mtime),
    )

    media = {**media, "provenance": get_provenance(f)}

    if not args.audio:
        try:
            has_sub = is_file_with_subtitle(f)
        except:
            has_sub = False
        media = {**media, "has_sub": has_sub}

    if args.audio:
        media = {**media, "listen_count": 0}

        try:
            tiny_tags = remove_None(TinyTag.get(f).as_dict())
        except:
            tiny_tags = dict()

        try:
            mutagen_tags = remove_None(mutagen.File(f).tags.as_dict())
        except:
            mutagen_tags = dict()

        tags = parse_tags(mutagen_tags, tiny_tags)
        return {**media, **tags}

    return media


def optimize_db(args):
    print("Optimizing database")
    if Path(args.db).exists():
        cmd("sqlite-utils", "optimize", args.db)
        columns = cmd(
            f"sqlite-utils tables {args.db} --columns | jq -r '.[0].columns[]'", quiet=True, shell=True
        ).stdout.splitlines()
        for column in columns:
            cmd("sqlite-utils", "create-index", "--if-not-exists", "--analyze", args.db, "media", column)


def extract_chunk(args, con, l):
    metadata = (
        Parallel(n_jobs=-1 if args.verbose == 0 else 1, backend="threading")(
            delayed(extract_metadata)(args, file) for file in l
        )
        or []
    )

    DF = pd.DataFrame(list(filter(None, metadata)))
    if args.audio:
        if DF.get(["year"]) is not None:
            DF.year = DF.year.astype(str)
    DF.apply(pd.to_numeric, errors="ignore").convert_dtypes().to_sql(  # type: ignore
        "media",
        con=con,
        if_exists="append",
        index=False,
        chunksize=70,
        method="multi",
    )


def find_new_files(args, con, path):
    video_files = get_video_files(path, args.audio)
    new_files = set(video_files)

    try:
        existing = set(
            map(
                lambda x: x["filename"],
                fetchall_dict(con, f"select filename from media where filename like '{path}%'"),
            )
        )
    except:
        video_files = list(new_files)
    else:
        video_files = list(new_files - existing)

        deleted_files = list(existing - new_files)
        if len(deleted_files) > 0:
            print(f"Removing {len(deleted_files)} orphaned metadata")

            df_chunked = chunks(deleted_files, SQLITE_PARAM_LIMIT)
            for l in df_chunked:
                con.execute(
                    "delete from media where filename in (" + ",".join(["?"] * len(l)) + ")",
                    (*l,),
                )
                con.commit()

        con.execute("delete from media where filename like '%/keep/%'")
        con.commit()

    return video_files


def scan_path(args, con, path):
    path = Path(path).resolve()
    print(f"{path} : Scanning...")

    video_files = find_new_files(args, con, path)

    if len(video_files) > 0:
        print(f"Adding {len(video_files)} new media")
        log.debug(video_files)

        batch_count = SQLITE_PARAM_LIMIT // 100
        chunks_count = math.ceil(len(video_files) / batch_count)
        df_chunked = chunks(video_files, batch_count)
        for idx, l in enumerate(df_chunked):
            percent = ((batch_count * idx) + len(l)) / len(video_files) * 100
            print(f"Extracting metadata: {percent:3.1f}% (chunk {idx + 1} of {chunks_count})")
            extract_chunk(args, con, l)

            if args.subtitle:
                print("Fetching subtitles")
                Parallel(n_jobs=5)(delayed(get_subtitle)(args, file) for file in l)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("db")
    parser.add_argument("paths", nargs="*")
    parser.add_argument("-a", "--audio", action="store_true")
    parser.add_argument("-s", "--subtitle", action="store_true")
    parser.add_argument("-yt", "--youtube-only", action="store_true")
    parser.add_argument("-sl", "--subliminal-only", action="store_true")
    parser.add_argument("-f", "--force-rescan", action="store_true")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    args = parser.parse_args()
    if args.force_rescan:
        Path(args.db).unlink(missing_ok=True)

    Path(args.db).touch()
    con = sqlite_con(args.db)
    for path in args.paths:
        scan_path(args, con, path)

    optimize_db(args)


if __name__ == "__main__":
    main()



import os
import sqlite3

from xklb.utils import stop


def sqlite_con(db):
    if not os.path.exists(db):
        print('Database file does not exist. Create one first with lb-extract.')
        stop()

    con = sqlite3.connect(db)
    con.row_factory = sqlite3.Row
    return con


def fetchall_dict(con, *args):
    return [dict(r) for r in con.execute(*args).fetchall()]
